from . import BaseTool
import logging
import json
logger = logging.getLogger('sigmund')


class BaseIDETool(BaseTool):
    """Base class for IDE tools that handles common command logic."""

    def __call__(self, **kwargs):
        """Generic call handler that builds command data from arguments."""
        command = self.__class__.__name__.replace('ide_', '')
        data = {"command": command}
        data.update(kwargs)
        # We don't ask for feedback directly, because OpenSesame will reply with
        # a user message.
        return ('(Suggesting IDE action)', json.dumps(data, indent=2),
                'json', False)

    @property
    def required_arguments(self):
        """Automatically derive required arguments from arguments dict."""
        return list(self.arguments.keys())


class ide_open_file(BaseIDETool):
    """Opens a file in the editor."""

    arguments = {
        'path': {
            'type': 'string',
            'description': 'The file path.'
        }
    }


class ide_execute_code(BaseIDETool):
    """Executes code in the editor. This code is executed locally on the user's system.'"""

    arguments = {
        'code': {
            'type': 'string',
            'description': 'The code to execute.'
        },
        'language': {
            'typ': 'string',
            'description': 'The programming language of the code.'
        }
    }



import os
import logging
logger = logging.getLogger('sigmund')
page_title = 'Sigmund AI'
ai_name = 'Sigmund'

# SERVER
#
# The external server address, that is, the URL that users visit
server_url = os.environ.get('FLASK_SERVER_URL', 'http://127.0.0.1:5000')
# The port at which the flask server is running internally. This can be
# different from the server URL if the app is running behind a proxy that
# redirects
flask_port = int(os.environ.get('FLASK_PORT', 5000))
# The flask host arhument where all 0s means listen to all incoming addresses
flask_host = os.environ.get('FLASK_HOST', '0.0.0.0')
# The secret key is used for logging in. This should be a long and arbitrary
# string that is hard to guess. This should not be shared
flask_secret_key = os.environ.get('FLASK_SECRET_KEY', '0123456789ABCDEF')
# A secret salt that is used to encrypt the messages on disk in combination
# with the user password
encryption_salt = os.environ.get('SIGMUND_ENCRYPTION_SALT', '0123456789ABCDEF').encode()

# PROMPT HISTORY
#
# The maximum length of a prompt in characters. If the prompt exceeds this 
# length, the start will be summarized.
max_prompt_length = 100000
# The length of the prompt to be summarized.
condense_chunk_length = 10000

# MESSAGES
# The maximum length of a user message
max_message_length = 10
# A fixed welcome message
welcome_message = '''Nice to meet you! I am Sigmund, your friendly AI assistant! How can I help you?'''
# The default title of a new conversation
default_conversation_title = 'New conversation'

# RATE LIMITS
#
# Maximum number of characters in a single message
max_message_length = 500_000
# The maximum number of tokens that can be consumed per hour by the answer
# model.
max_tokens_per_hour = 1_000_000
max_tokens_per_hour_exceeded_message = 'You have reached the hourly usage limit. Please wait and try again later!'
anthropic_max_thinking_tokens = 2048

# LOGGING
#
# When set to True, replies will be logged. This should disabled during 
# production for privacy.
log_replies = os.environ.get('SIGMUND_LOG_REPLIES', False)

# MODELS
#
# The API keys should not be shared
openai_api_key = os.environ.get('OPENAI_API_KEY', None)
anthropic_api_key = os.environ.get('ANTHROPIC_API_KEY', None)
mistral_api_key = os.environ.get('MISTRAL_API_KEY', None)
bfl_api_key = os.environ.get('BFL_API_KEY', None)
# Force dummy model for development purposes
dummy_model = int(os.environ.get("SIGMUND_DUMMY_MODEL", False))
# Supported models are continuously updated. See sigmund.model for more
# information.
# Model configuations are combinations of models that can be used together
# - The search model is used to formulate search queries and evaluate whether
#   documents are relevant. This can be a cheap model.
# - The condense model is used to summarize message history when the 
#   conversation becomes too long. This can also be a cheap model.
# - The answermodel generates the actual answer. This should be a very capable
#   model
model_config = {
    'openai': {
        'condense_model': 'gpt-5-mini',
        'public_model': 'gpt-5-nano',
        'answer_model': 'gpt-5.2'
    },
    'openai_thinking': {
        'condense_model': 'gpt-5-mini',
        'public_model': 'gpt-5-nano',
        'answer_model': 'gpt-5.2-thinking'
    },
    'anthropic': {
        'condense_model': 'claude-4-5-haiku',
        'public_model': 'claude-4-5-haiku',
        'answer_model': 'claude-4-6-sonnet'
    },
    'anthropic_thinking': {
        'condense_model': 'claude-4-5-haiku',
        'public_model': 'claude-4-5-haiku',
        'answer_model': 'claude-4-6-opus-thinking'
    },
    'mistral': {
        'condense_model': 'mistral-small-latest',
        'public_model': 'mistral-small-latest',
        'answer_model': 'mistral-large-latest',
        'vision_model': 'mistral-large-latest'
    },
    'dummy': {
        'condense_model': 'dummy',
        'public_model': 'dummy',
        'answer_model': 'dummy'
    }
}
# Model-specific keyword arguments that are passed to the message generation
# functions
anthropic_kwargs = {
    'max_tokens': 65536
}
openai_kwargs = {}
mistral_kwargs = {}

# TOOLS
#
tools = ['search_google_scholar', 'search_openalex', 'download_from_openalex',
         'generate_image_dalle3', 'generate_image_flux',
         'opensesame_select_item', 'opensesame_new_item',
         'opensesame_remove_item_from_parent', 'opensesame_rename_item',
          'opensesame_add_existing_item_to_parent',
         'opensesame_update_item_script', 'opensesame_set_global_var',
         'ide_execute_code', 'ide_open_file']

# SETTINGS
#
# These are default values for settings, which are stored in the database.
settings_default = {
    # Indicates the model configuration as specified above
    'model_config': 'openai',
    # Indicates which tools are enabled
    'tool_search_google_scholar': 'false',
    'tool_search_openalex': 'false',
    'tool_download_from_openalex': 'false',
    'tool_generate_image_dalle3': 'false',
    'tool_generate_image_flux': 'false',
    'tool_opensesame_select_item': 'false',
    'tool_opensesame_new_item': 'false',
    'tool_opensesame_remove_item_from_parent': 'false',
    'tool_opensesame_rename_item': 'false',
    'tool_opensesame_add_existing_item_to_parent': 'false',
    'tool_opensesame_update_item_script': 'false',
    'tool_opensesame_set_global_var': 'false',
    'tool_ide_execute_code': 'false',
    'tool_ide_open_file': 'false',
    # Indicates which library collections are enabled
    'collection_opensesame': 'true',
    'collection_datamatrix': 'false',
    'collection_forum': 'false'
}


# OPENALEX
openalex_api_key = os.environ.get('OPENALEX_API_KEY', None)

# DOCUMENTATION
#

# Enables library search
search_enabled = True
# The library is organized into collections
search_collections = {'opensesame', 'datamatrix', 'forum'}
# The maximum distance should be determined empirically by comparing distances
# between realistic queries and relevant document, and unrelated queries and the
# documentation.
search_max_distance = 1.15
# If no results are found, we try again with a higher threshold. This may be
# necessary especially for short questions.
search_max_distance_fallback = 1.35
# The minimum length of the search query. If the query is shorter, we previous
# messages are prepended until the query reaches the minimum length or there are
# no more messages.
search_min_query_length = 1000
# We also have a maximum query length, because the embedding model has a maximum
# number of tokens. We truncate longer queries.
search_max_query_length = 10000
# The number of documents. This applies separately to regular documentation and
# howtos, and does not include foundation documents.
search_docs_max = 2
# The embedding model used for search. This has the name of an OpenAI embedding
# model
search_embedding_model = 'text-embedding-3-large'
# The number of documents and maximum number of words that is kept for public 
# search
public_search_docs_max = 6
public_search_max_doc_length = 1000
search_persist_directory = "library"
search_embedding_provider = "openai"
search_embedding_model = "text-embedding-3-large"



def validate_user(username, password):
    """user_validation.validate() should connect to an authentication system
    that verifies the account. 
    
    Parameters
    ----------
    username : str
        Lowercase. Whitespace has been stripped.
    password : str
        Whitespace has been stripped.
        
    Returns
    -------
    str or None:
        If the user is validated, the username is returned. This is typically
        equal to the `username` parameter, but in some cases it may differ, for
        example if the user logs in with an email address instead of the actual
        username. If the user is not validated, None is returned.
    """
    try:
        import user_validation
    except ImportError:
        logger.info('no user validation script found')
        return username
    logger.info('using validation script')
    return user_validation.validate(username, password)


# SUBSCRIPTIONS
#
# Enable this to activate the Stripe-based subscription functionality.
subscription_required = bool(
    int(os.environ.get('SIGMUND_SUBSCRIPTION_REQUIRED', 0)))
# This is the duration of the subscription in days. This should be set to a bit
# longer than a month to provide a grace period in case of payment issues.
subscription_length = 50
#
# Stripe is a payment portal. The information below should not be shared
#
# The Price ID identifies the product. You can find it in the product catalog
# of the Stripe dashboard.
stripe_price_id = os.environ.get('STRIPE_PRICE_ID', None)
# The publishable key is the API keys section of the Developers dashboard
stripe_publishable_key = os.environ.get('STRIPE_PUBLISHABLE_KEY', None)
# The secret (API) key is the API keys section of the Developers dashboard
stripe_secret_key = os.environ.get('STRIPE_SECRET_KEY', None)
# The webhook secret is used to ensure that webhook calls actually come from
# stripe
stripe_webhook_secret = os.environ.get('STRIPE_WEBHOOK_SECRET', None)
# Additional keywords passed to stripe.checkout.Session.create to customize
# the checkout process
stripe_checkout_keywords = dict(
    allow_promotion_codes=True,
    payment_method_types=['card', 'ideal', 'bancontact', 'paypal'])


# GOOGLE LOGIN
#
google_login_enabled = True
google_client_id = os.environ.get("GOOGLE_CLIENT_ID", None)
google_client_secret = os.environ.get("GOOGLE_CLIENT_SECRET", None)
google_discovery_url = (
    "https://accounts.google.com/.well-known/openid-configuration"
)



import logging
import uuid
from cryptography.fernet import InvalidToken
import multiprocessing as mp
from . import prompt, config, utils
logger = logging.getLogger('sigmund')


class Messages:
    
    def __init__(self, sigmund, persistent=False):
        self._sigmund = sigmund
        self._persistent = persistent
        self._conversation_id = None
        self.workspace_content = None
        self.workspace_language = None
        if self._persistent:
            self.load()
        else:
            self.init_conversation()
        
    def __len__(self):
        return len(self._message_history)
        
    def __iter__(self):
        for msg in self._message_history:
            yield msg
            
    def init_conversation(self):
        self._condensed_text = None
        metadata = self.metadata()
        metadata['answer_model'] = 'welcome-bot'
        self._conversation_title = config.default_conversation_title
        self._message_history = [['assistant', self.welcome_message(), 
                                  metadata]]
        self._condensed_message_history = [
            [role, content] for role, content, metadata
            in self._message_history[:]]        
            
    def metadata(self, workspace_content: str = None,
                 workspace_language: str = None, message_id: str = None):
        return {'message_id': str(uuid.uuid4()) 
                    if message_id is None else message_id,
                'workspace_content': workspace_content,
                'workspace_language': workspace_language,
                'timestamp': utils.current_datetime(),
                'sources': self._sigmund.documentation.to_json(),
                'condense_model': self._sigmund.model_config['condense_model'],
                'answer_model': self._sigmund.model_config['answer_model']}
        
    def append(self, role: str, message: str, workspace_content: str = None,
               workspace_language: str = None, message_id: str = None):
        metadata = self.metadata(workspace_content=workspace_content,
                                 workspace_language=workspace_language,
                                 message_id=message_id)
        self._message_history.append([role, message, metadata])
        self._condensed_message_history.append([role, message])
        self._condense_message_history()
        if self._persistent:
            self.save()
        return metadata
    
    def delete(self, message_id):
        message_to_remove = None
        condensed_message_to_remove = None
        for role, message, metadata in self._message_history:
            if metadata['message_id'] == message_id:
                message_to_remove = [role, message, metadata]
                condensed_message_to_remove = [role, message]
                break
        else:
            logger.info(f'message not found for deletion: {message_id}')
            return
        if message_to_remove:
            logger.info(f'deleting message: {message_id}')
            self._message_history.remove(message_to_remove)
        if condensed_message_to_remove:
            try:
                self._condensed_message_history.remove(
                    condensed_message_to_remove)
            except ValueError:
                logger.error(
                    f'Could not find condensed message to remove for {message_id}')
        if self._persistent:
            self.save()
            
    def prompt(self, system_prompt=None):
        """The prompt consists of the system prompt followed by a sequence of
        AI, user, and tool/ function messages.
        
        If no system prompt is provided, one is automatically constructed.
        Typically, an explicit system_prompt is provided during the search
        phase, but not during the answer phase.
        """
        if system_prompt is None:
            system_prompt = self._system_prompt()
        model_prompt = [dict(role='system', content=system_prompt)]
        for msg_nr, (role, content) in enumerate(
                self._condensed_message_history):
            content = utils.remove_masked_elements(content)
            if role == 'user':
                # Prefix the last message with the current workspace
                if self.workspace_content and \
                        msg_nr == len(self._condensed_message_history) - 1:
                    content = prompt.render(
                        prompt.CURRENT_WORKSPACE,
                        workspace_content=self.workspace_content,
                        workspace_language=self.workspace_language) + content
            model_prompt.append(dict(role=role, content=content))
        return model_prompt
    
    def visible_messages(self):
        """Yields role, message, metadata while ignoring messages and 
        converting tool messages into user messages with tool result as 
        content. This is mainly for display in the web interface.
        """
        for role, message, metadata in self:
            if role == 'tool':
                continue
            if not message.strip():
                continue
            yield role, message, metadata

    def welcome_message(self):
        return config.welcome_message
        
    def _condense_message_history(self):
        system_prompt = self._system_prompt()
        # Determine the effective length of the condensed prompt while
        # excluding masked content. Makes content corresponds for example to
        # image data, and other kinds of data that are shown to the user but not
        # passed to the model.
        prompt_length = sum([
            len(utils.remove_masked_elements(content))
            for role, content in self._condensed_message_history
        ])
        logger.info(f'system prompt length: {len(system_prompt)}')
        logger.info(
            f'condensed prompt length (without system prompt): {prompt_length}')
        if prompt_length <= config.max_prompt_length:
            logger.info('no need to condense')
            return
        condense_length = 0
        condense_messages = []
        while self._condensed_message_history and \
                condense_length < config.condense_chunk_length:
            if len(self._condensed_message_history) <= 2:
                break
            role, content = self._condensed_message_history.pop(0)
            content = utils.remove_masked_elements(content)
            condense_length += len(content)
            condense_messages.insert(0, (role, content))
        logger.info(f'condensing {len(condense_messages)} messages')
        condense_prompt = prompt.render(
            prompt.CONDENSE_HISTORY,
            history=''.join(
                f'You said: {content}' if role == 'system' 
                else f'User said: {content}'
                for role, content in condense_messages))
        result = self._sigmund.condense_model.predict(condense_prompt)
        # Just in case a model gets confused and returns a tool call.
        if not isinstance(result, str):
            logger.error(f'condense model returned non-string result: {result}')
            return
        self._condensed_text = result.strip()
        
    def _system_prompt(self):
        """The system prompt that is used for question answering consists of
        several fragments.
        """
        system_prompt = [prompt.SYSTEM_PROMPT_IDENTITY]
        if self._sigmund.transient_system_prompt:
            system_prompt.append(self._sigmund.transient_system_prompt)
        # If available, documentation is also included in the prompt
        if len(self._sigmund.documentation):
            system_prompt.append(self._sigmund.documentation.prompt())
        # And finally, if the message history has been condensed, this is also
        # included.
        if self._condensed_text:
            logger.info('appending condensed text to system prompt')
            system_prompt.append(prompt.render(
                prompt.SYSTEM_PROMPT_CONDENSED,
                summary=self._condensed_text))
        # Combine all non-empty prompt chunks
        return '\n\n'.join(chunk for chunk in system_prompt if chunk.strip())    

    def load(self):
        try:
            conversation = self._sigmund.database.get_active_conversation()
        except InvalidToken as e:
            self.init_conversation()
            self.save()
            logger.error(f'failed to get active conversation: {e}')
            return
        if not conversation['message_history']:
            self.init_conversation()
            return
        modified = False
        for _, _, metadata in conversation['message_history']:
            if 'message_id' not in metadata:
                metadata['message_id'] = str(uuid.uuid4())
                modified = True
        self._conversation_title = conversation['title']
        self._message_history = conversation['message_history']
        self._condensed_text = conversation['condensed_text']
        self._condensed_message_history = \
            conversation['condensed_message_history']
        if self._persistent and modified:
            self.save()
    
    def save(self):
        conversation = {
            'condensed_text': self._condensed_text,
            'message_history': self._message_history,
            'condensed_message_history': self._condensed_message_history,
            'title': self._conversation_title,
        }
        self._conversation_id = self._sigmund.database.update_active_conversation(conversation)
        # We update the title in a background process so that we don't block
        # the conversation
        mp.Process(target=self._update_title).start()

    def _update_title(self):
        """The conversation title is updated when there are at least two 
        messages, excluding the system prompt and AI welcome message. Based on
        the last messages, a summary title is then created.
    
        This method is run in a separate process to avoid blocking the main thread.
        """
        if len(self) <= 2 or \
                self._conversation_title != config.default_conversation_title:
            return
        logger.info('updating conversation title')
        title_prompt = [dict(role='system', content=prompt.TITLE_PROMPT)]
        title_prompt += self.prompt()[2:]
        suggested_title = self._sigmund.condense_model.predict(title_prompt)
        # The prediction may be a tool call, so we need to check if it is a str.
        # This should not ordinarily happen, but sometimes models get confused.
        if not isinstance(suggested_title, str):
            logger.error(f'suggested conversation title is not str, but: {suggested_title}')
            return
        self._conversation_title = suggested_title.strip('"\'')
        if len(self._conversation_title) > 100:
            self._conversation_title = self._conversation_title[:100] + '…'
        self._sigmund.database.set_conversation_title(
            self._conversation_id, self._conversation_title)
        logger.info('completed updating conversation title')



import json
import logging
import time
from datetime import datetime, timedelta
from .. import config
from .models import db, User, Conversation, Activity, Subscription, Setting, \
    Message
from .encryption import EncryptionManager
from sqlalchemy import func, delete
from sqlalchemy.orm.exc import NoResultFound

logger = logging.getLogger('sigmund')


class DatabaseManager:
    
    def __init__(self, sigmund, username: str,
                 encryption_key: [str, bytes]=None):
        self.transient_settings = {}
        self._sigmund = sigmund
        self.username = username
        self.encryption_manager = EncryptionManager(encryption_key)
        self.ensure_user_exists()
        logger.info(
            f'initializing database for user {self.username} ({self.user_id})')
        
    def prune_detached_messages(self):
        """Delete all messages that are not attached to a conversation anymore
        due to an old bug in which messages where re-added but never deleted.
        """
        for conversation in Conversation.query.filter_by(
                user_id=self.user_id).all():
            logger.info(f'decrypting {conversation.conversation_id}, {len(conversation.data)} bytes')
            if len(conversation.data) > 20000:
                logger.info('skipping')
                continue
            decrypted_data = self.encryption_manager.decrypt_data(
                conversation.data)
            conversation_data = json.loads(decrypted_data)
            logger.info('getting attached messages')
            attached_message_ids = conversation_data.get('message_history', [])
            attached_message_ids = [
                msg_id for msg_id in attached_message_ids
                if isinstance(msg_id, int)
            ]
            logger.info('getting all messages')
            all_messages = Message.query.filter_by(
                conversation_id=conversation.conversation_id).all()
            for message in all_messages:
                if message.message_id not in attached_message_ids:
                    logger.info(f'deleting {message.message_id}')
                    db.session.delete(message)
        logger.info('committing')
        db.session.commit()
        logger.info('done')
            
    def ensure_user_exists(self):
        try:
            user = User.query.filter_by(username=self.username).one()
            self.user_id = user.user_id
            logger.info(f'user {self.username} exists')
        except NoResultFound:
            # This is a hack to fix a bug in earlier versions of Sigmund without
            # breaking log-in for existing users.
            #
            # If the user id doesn't exist, this may be because the user 
            # previously logged in through Google before the unique ID was 
            # included in the username. Currently, Google usernames look like
            # Name (google)::unique_id. Previously, Google usernames looked like
            # Name (google). If an old Google username exists, we automatically
            # rename the user to the new unique username.
            if '(google)::' in self.username:
                # Strip everything after (google)::
                old_username = self.username.split('(google)::')[0] + '(google)'
                logger.info(
                    f'Checking if {self.username} exists as {old_username}')
                try:                    
                    user = User.query.filter_by(username=old_username).one()
                    self.user_id = user.user_id
                except NoResultFound:
                    # No the user really doesn't exist
                    pass
                else:
                    # Yes, the user has an old-style name, so we need to update
                    # the database so that User.username is set to old username
                    user.username = self.username
                    db.session.commit()
                    return
            # If the user doesn't exist, we create it and start a new 
            # conversation
            logger.info(f'creating new user: {self.username}')
            user = User(username=self.username)
            db.session.add(user)
            db.session.commit()
            self.user_id = user.user_id
            self.new_conversation()
    
    def get_message(self, message_id: int) -> dict:
        # If it's not an int, then it's an old-style mesage
        if not isinstance(message_id, int):
            return message_id
        try:
            # First, check if the message exists in the new Message table
            message = Message.query.filter_by(message_id=message_id).first()
            if message:
                # If it exists, decrypt and return the message
                decrypted_data = self.encryption_manager.decrypt_data(message.data)
                return json.loads(decrypted_data)
            # If the message doesn't exist, something is wrong.
            logger.error(f"Message {message_id} does not exist")
            return {}
        except Exception as e:
            logger.error(f"Error retrieving message {message_id}: {e}")
            return {}
            
    def get_message_history(self, conversation_data):
        """Returns a list of message dicts, excluding empty ones"""
        message_ids = conversation_data.get('message_history', [])
        message_history = []
        for msg_id in message_ids:
            msg = self.get_message(msg_id)
            if msg:
                message_history.append(msg)
        return message_history

    
    def add_message(self, conversation_id: int, message_data: dict) -> int:
        json_data = json.dumps(message_data)
        encrypted_data = self.encryption_manager.encrypt_data(
            json_data.encode('utf-8'))
        message = Message(conversation_id=conversation_id, data=encrypted_data)
        db.session.add(message)
        db.session.commit()
        return message.message_id
    
    def get_active_conversation(self) -> dict:
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=user.active_conversation_id).one()
        except NoResultFound:
            logger.warning(f"No active conversation found for user {self.user_id}")
            return {}    
        decrypted_data = self.encryption_manager.decrypt_data(conversation.data)
        conversation_data = json.loads(decrypted_data)
        conversation_data['message_history'] = self.get_message_history(
            conversation_data)
        return conversation_data            

    def _get_conversation(self, conversation_id: int):
        """Retrieves a conversation for the current user.

        Parameters
        ----------
        conversation_id : int
            The ID of the conversation to retrieve

        Returns
        -------
        Conversation or None
            The Conversation object if found and belongs to the user, None otherwise
        """
        try:
            conversation = Conversation.query.filter_by(
                conversation_id=conversation_id, user_id=self.user_id).one()
            return conversation
        except NoResultFound:
            logger.warning(f"Conversation {conversation_id} not found or does "
                           f"not belong to user {self.user_id}")
            return None

    def _decrypt_conversation_data(self, conversation):
        """Decrypts and parses conversation data.

        Parameters
        ----------
        conversation : Conversation
            The conversation object with encrypted data

        Returns
        -------
        dict
            The decrypted and parsed conversation data as a dictionary
        """
        decrypted_data = self.encryption_manager.decrypt_data(conversation.data)
        return json.loads(decrypted_data)

    def _encrypt_conversation_data(self, conversation_data: dict):
        """Encrypts conversation data.

        Parameters
        ----------
        conversation_data : dict
            The conversation data dictionary to encrypt

        Returns
        -------
        bytes
            The encrypted data
        """
        json_data = json.dumps(conversation_data)
        return self.encryption_manager.encrypt_data(json_data.encode('utf-8'))

    def _update_conversation_data(self, conversation, conversation_data: dict):
        """Updates and saves encrypted conversation data.

        Parameters
        ----------
        conversation : Conversation
            The conversation object to update
        conversation_data : dict
            The updated conversation data dictionary
        """
        encrypted_data = self._encrypt_conversation_data(conversation_data)
        conversation.data = encrypted_data
        db.session.commit()

    def set_active_conversation(self, conversation_id: int) -> bool:
        """Sets the active conversation for the current user.

        This method retrieves the specified conversation, updates its last_updated
        timestamp, encrypts the updated data, and sets it as the user's active
        conversation.

        Parameters
        ----------
        conversation_id : int
            The ID of the conversation to set as active

        Returns
        -------
        bool
            True if the conversation was successfully set as active, False if
            the conversation was not found or doesn't belong to the user
        """
        conversation = self._get_conversation(conversation_id)
        if conversation is None:
            return False
    
        # Update the last_updated timestamp
        conversation_data = self._decrypt_conversation_data(conversation)
        conversation_data['last_updated'] = time.time()
        self._update_conversation_data(conversation, conversation_data)
    
        # Change the active conversation for the user
        user = User.query.filter_by(user_id=self.user_id).one()
        user.active_conversation_id = conversation.conversation_id
        db.session.commit()
        return True

    def set_conversation_title(self, conversation_id: int, title: str) -> bool:
        """Updates the title of a conversation.

        This method retrieves the specified conversation, decrypts its data,
        updates the title field, and re-encrypts the data before committing
        to the database.

        Parameters
        ----------
        conversation_id : int
            The ID of the conversation to update
        title : str
            The new title for the conversation

        Returns
        -------
        bool
            True if the title was successfully updated, False if the
            conversation was not found or doesn't belong to the user
        """
        conversation = self._get_conversation(conversation_id)
        if conversation is None:
            return False
    
        # Update the title
        conversation_data = self._decrypt_conversation_data(conversation)
        conversation_data['title'] = title
        self._update_conversation_data(conversation, conversation_data)
        return True
    
    def update_active_conversation(self, conversation_data: dict) -> bool:
        """Updates the active conversation. The messages are saved in a 
        separate table as separate entries. However, all messages are replaced
        when the conversation is updated. That the conversation doesn't grow
        when a new message is added, but rather is reconstructed.
        """
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=user.active_conversation_id
            ).one()
        except NoResultFound:
            logger.warning(
                f"No active conversation to update for user {self.user_id}")
            return            
        logger.info(f'updating conversation {user.active_conversation_id}')
        logger.info('deleting old messages')
        # Batch delete all messages linked to this conversation, because we
        # will recreate new messages
        db.session.execute(
            delete(Message).where(
                Message.conversation_id == conversation.conversation_id)
        )
        logger.info('adding new messages')
        # Extract new messages and store them separately
        message_history = conversation_data.pop('message_history', [])
        message_ids = []
        for message in message_history:
            message_id = self.add_message(conversation.conversation_id,
                                          message)
            message_ids.append(message_id)
        conversation_data['message_history'] = message_ids
        conversation_data['last_updated'] = time.time()
        json_data = json.dumps(conversation_data)
        encrypted_data = self.encryption_manager.encrypt_data(
            json_data.encode('utf-8'))
        conversation.data = encrypted_data
        logger.info('committing conversation')
        db.session.commit()
        logger.info('done')
        return conversation.conversation_id

    def list_conversations(self, query=None) -> dict:
        conversations = {}
        user = User.query.filter_by(user_id=self.user_id).one()
        for conversation in \
                Conversation.query.filter_by(user_id=self.user_id).all():
            try:
                data = json.loads(self.encryption_manager.decrypt_data(
                    conversation.data))
                # Don't list empty conversations except for the active one
                if len(data.get('message_history', [])) < 2 and \
                        user.active_conversation_id != conversation.conversation_id:
                    continue
                title = data.get('title', 'Untitled conversation')
                if query is not None:
                    # First check title match
                    if query.lower() not in title.lower():
                        # If the title doesn't match, then we do a full-text
                        # search on the message history
                        match = False
                        for message in self.get_message_history(data):
                            message_text = message[1]
                            if query.lower() in message_text.lower():
                                match = True
                                break
                        if not match:
                            continue
                conversations[conversation.conversation_id] = (
                     title, data.get('last_updated', time.time()))
            except Exception as e:
                logger.error(f"Error decrypting conversation data: {e}")
        return conversations
    
    def export_conversations(self) -> dict:
        conversations = []
        for conversation in \
                Conversation.query.filter_by(user_id=self.user_id).all():
            try:
                decrypted_data = self.encryption_manager.decrypt_data(
                    conversation.data)
                conversation_data = json.loads(decrypted_data)
                message_history = self.get_message_history(conversation_data)
            except Exception as e:
                logger.error(f"Error decrypting conversation data: {e}")
            else:
                conversation_data['message_history'] = message_history
                conversations.append(conversation_data)
        return conversations    

    def new_conversation(self) -> bool:
        try:
            conversation_data = {
                "title": "New conversation",
                "condensed_text": None,
                "message_history": [],
                "message_metadata": [],
                "last_updated": time.time()
            }
            json_data = json.dumps(conversation_data)
            encrypted_data = self.encryption_manager.encrypt_data(
                json_data.encode('utf-8'))
            conversation = Conversation(user_id=self.user_id,
                                        data=encrypted_data)
            db.session.add(conversation)
            db.session.commit()
            self.set_active_conversation(conversation.conversation_id)
            return True
        except Exception as e:
            logger.error(f"Error creating new conversation: {e}")
            return False
    
    def delete_conversation(self, conversation_id: int) -> bool:
        try:
            user = User.query.filter_by(user_id=self.user_id).one()
            conversation = Conversation.query.filter_by(
                conversation_id=conversation_id, user_id=self.user_id).one()
            if user.active_conversation_id == conversation.conversation_id:
                logger.warning(
                    f"Cannot delete active conversation {conversation_id} for "
                    f"user {self.user_id}")
                return False

            # Delete associated messages
            Message.query.filter_by(conversation_id=conversation_id).delete()

            # Delete the conversation
            db.session.delete(conversation)
            db.session.commit()
            return True
        except NoResultFound:
            logger.warning(f"Conversation {conversation_id} not found or does "
                           f"not belong to user {self.user_id}")
            return False

    def add_activity(self, tokens_consumed: int):
        """Adds the number of tokens consumed using the current time."""
        new_activity = Activity(user_id=self.user_id, time=datetime.utcnow(),
                                tokens_consumed=tokens_consumed)
        db.session.add(new_activity)
        db.session.commit()

    def get_activity(self, after_time=None) -> int:
        """Retrieves the total number of tokens consumed since a particular 
        time, using the past hour as a default if no after_time is provided.
        """
        if after_time is None:
            after_time = datetime.utcnow() - timedelta(hours=1)

        total_tokens = db.session.query(func.sum(Activity.tokens_consumed)) \
            .filter(Activity.user_id == self.user_id) \
            .filter(Activity.time >= after_time) \
            .scalar()
        return total_tokens if total_tokens is not None else 0

    def add_subscription_record(self, stripe_customer_id: str,
                                stripe_subscription_id: str,
                                from_date=None, to_date=None):
        """Always append a new subscription record. Do not update old rows.
        from_date defaults to now; to_date defaults to now + subscription_length.
        """
        now = datetime.utcnow()
        from_date = from_date or now
        to_date = to_date or now + timedelta(days=config.subscription_length)
        subscription = Subscription(
            user_id=self.user_id,
            from_date=from_date,
            to_date=to_date,
            stripe_customer_id=stripe_customer_id,
            stripe_subscription_id=stripe_subscription_id
        )
        db.session.add(subscription)
        db.session.commit()

    def update_subscription(self, stripe_customer_id: str,
                            stripe_subscription_id: str, from_date=None,
                            to_date=None):
        """Backwards-compatible wrapper that appends a new subscription record."""
        return self.add_subscription_record(
            stripe_customer_id=stripe_customer_id,
            stripe_subscription_id=stripe_subscription_id,
            from_date=from_date,
            to_date=to_date
        )

    def check_subscription(self) -> bool:
        """Return True if at least one subscription row is currently active."""
        now = datetime.utcnow()
        # Efficient existence check
        return db.session.query(
            db.exists().where(
                (Subscription.user_id == self.user_id) &
                (Subscription.from_date <= now) &
                (Subscription.to_date > now)
            )
        ).scalar()

    def get_stripe_customer_id(self) -> str:
        """Return the stripe_customer_id of the subscription with the latest to_date.
        If multiple share the same to_date, pick the latest from_date as tie-breaker.
        """
        subscription_record = (
            Subscription.query
            .filter(Subscription.user_id == self.user_id)
            .order_by(Subscription.to_date.desc(), Subscription.from_date.desc())
            .first()
        )
        return subscription_record.stripe_customer_id if subscription_record else None

    @staticmethod
    def from_stripe_customer_id(stripe_customer_id: str):
        """Return a DatabaseManager instance for the user most recently associated
        with this stripe_customer_id (based on latest to_date, then from_date).
        """
        subscription_record = (
            Subscription.query
            .filter(Subscription.stripe_customer_id == stripe_customer_id)
            .order_by(Subscription.to_date.desc(), Subscription.from_date.desc())
            .first()
        )
        if not subscription_record:
            return None
        user_record = User.query.filter(User.user_id == subscription_record.user_id).first()
        if not user_record:
            return None
        return DatabaseManager(None, username=user_record.username)

    def get_setting(self, key: str) -> str:
        """Retrieve a setting value for the current user, which is available
        as self.user_id. If the setting does not exist, return the default
        value as specified in the config or None if no default has been 
        specified.
        """
        if key in self.transient_settings:
            return self.transient_settings[key]
        setting = db.session.query(Setting).filter_by(
            user_id=self.user_id, key=key).first()
        return setting.value if setting \
            else config.settings_default.get(key, None)

    def set_setting(self, key: str, value: str, transient: bool = False):
        """Set a setting to specified value for the current user. If the
        setting already exists, overwrite it. When transient is True, the 
        setting is not stored in the database, but only used for this session.
        """
        if transient:
            self.transient_settings[key] = value
            return
        setting = Setting.query.filter_by(
            user_id=self.user_id, key=key).first()
        if setting:
            setting.value = value
        else:
            new_setting = Setting(user_id=self.user_id, key=key,
                                  value=value)
            db.session.add(new_setting)
        db.session.commit()

    def list_settings(self) -> dict:
        """Returns a dict of all settings for the current user."""
        settings = {}
        setting_objs = db.session.query(Setting).filter_by(
            user_id=self.user_id).all()
        for setting in setting_objs:
            settings[setting.key] = setting.value
        settings.update(self.transient_settings)
        return settings



import logging
import json
from types import GeneratorType
from . import config
from .reply import Reply, ActionReply
from .documentation import Documentation
from .messages import Messages
from .model import model
from . import tools as mod_tools
from .database.manager import DatabaseManager
from . import utils
logger = logging.getLogger('sigmund')


class Sigmund:
    """The main chatbot class.
    
    Parameters
    ----------
    user_id: A user name
    persistent: Indicates whether the current session should be persistent in
        the sense of using the database.
    encryption_key: A key to encrypt the database.
    model_config: Indicates the model configuration. If none, the config
        default will be used.
    tools: A list of tools to use. If None, the config default will be used.
    transient_settings: A dictionary of settings that will be used for the
        current session only.
    transient_system_prompt: A system prompt that will be added to the default
        system prompt for the current session only.
    foundation_document_topics: A list of topics. Foundation documents matching 
        these topics are always included in the context.
    """
    def __init__(self, user_id: str, persistent: bool = False,
                 encryption_key: str = None,
                 model_config: str = None,
                 tools: list = None,
                 transient_settings: dict = None,
                 transient_system_prompt: str = None,
                 foundation_document_topics: list = None):
        self.user_id = user_id
        self.database = DatabaseManager(self, user_id, encryption_key)
        if transient_settings:
            logger.info(f'using transient settings: {transient_settings}')
            self.database.transient_settings = transient_settings
        self.transient_system_prompt = transient_system_prompt
        # Available model configs may change with updates, but the user settings
        # are not updated along with this. Therefore, if a model config doesn't
        # exist, we reset to the default.
        if model_config is None:
            model_config = self.database.get_setting('model_config')
        if model_config not in config.model_config:
            logger.warning(
                f'model_config {model_config} not found, resetting to default')
            model_config = config.settings_default['model_config']
            self.database.set_setting('model_config', model_config)
        self.model_config = config.model_config[model_config]
        self.documentation = Documentation(
            self, foundation_document_topics=foundation_document_topics)
        self.messages = Messages(self, persistent)
        if tools is None:
            tools = [t for t in config.tools
                     if self.database.get_setting(f'tool_{t}') == 'true']
        # Tools are class names from the tools module, which need to be
        # instantiated with sigmund (self) as first argument
        self.tools = [getattr(mod_tools, t)(self) for t in tools]
        # If there are answer tools, the mode can choose freely
        if tools:
            tool_choice = 'auto'
        else:
            tool_choice = None
        self.answer_model = model(self, self.model_config['answer_model'],
                                  tools=self.tools,
                                  tool_choice=tool_choice)
        self.condense_model = model(self, self.model_config['condense_model'])
        self.public_model = model(self, self.model_config['public_model'])
    
    def send_user_message(self, message: str, workspace_content: str = None,
                          workspace_language: str = 'text',
                          attachments: list = None,
                          message_id: str = None) -> GeneratorType:
        """The main function that takes a user message and returns one or 
        replies. This is a generator function where each yield gives a tuple.
        
        This tuple can be (dict, dict, str) in which case the first dict 
        contains an action and a message key. This communicates to the client 
        that an action should be taking, such that the loading indicator should
        change. This tuple can also be (str, dict, str) in which case the first
        str contains an AI message and the dict contains metadata for the 
        message. The third element corresponds to the content of the workspace,
        and can be None if empty.
        """
        # A temporary hack to prune messages for accounts
        if message == 'prune_detached_messages()':
            yield ActionReply(
                f'{config.ai_name} is pruning detached messages ')
            self.database.prune_detached_messages()
            return Reply('Done', None, None, None)
        if config.log_replies:
            logger.info(f'[user message] {message}')
        if self._rate_limit_exceeded():
            yield Reply(config.max_tokens_per_hour_exceeded_message,
                        self.messages.metadata())
            return
        self.messages.workspace_content = workspace_content
        self.messages.workspace_language = workspace_language
        self.messages.append('user', 
                             message=message,
                             workspace_content=workspace_content,
                             workspace_language=workspace_language,
                             message_id=message_id)
        if config.search_enabled and self.documentation.enabled:
            # Search queries work best if they'r not too short. That's why we
            # include preceding messages until the search query reaches the
            # minimum length, or until there are no more messages. (The welcome
            # message is always excluded.)
            query_messages = []
            query_len = 0
            # Inversely loop through the message history, ignoring the very first
            # message, because it is a welcome message.
            for _, msg, _ in self.messages._message_history[-1:0:-1]:
                query_messages.insert(0, msg)
                query_len += len(msg)
                if query_len > config.search_min_query_length:
                    break
            logger.info(f'using {len(query_messages)} message(s) for search query')
            logger.info(f'query length: {query_len}')
            query = '\n\n'.join(query_messages)
            query = query[:config.search_max_query_length]
            for reply in self._search(query):
                yield reply
        for reply in self._answer(attachments):
            yield reply
            
    def _rate_limit_exceeded(self):
        tokens_consumed_past_hour = self.database.get_activity()
        logger.info(
            f'tokens consumed in past hour: {tokens_consumed_past_hour}')
        return tokens_consumed_past_hour > config.max_tokens_per_hour
    
    def _search(self, message: str) -> GeneratorType:
        """Implements the documentation search phase."""
        yield ActionReply(f'{config.ai_name} is searching ')
        logger.info('[search state] entering')
        self.documentation.clear()
        # First seach based on the user question
        logger.info('[search state] searching based on user message')
        self.documentation.search(message)
        logger.info(
            f'[search state] {len(self.documentation._documents)} documents, {len(self.documentation)} characters')
    
    def _answer(self, attachments: [] = None,
                state: str = 'answer') -> GeneratorType:
        """Implements the answer phase."""
        yield ActionReply(f'{config.ai_name} is thinking and typing ')
        logger.info(f'[{state} state] entering')
        # We first collect a regular reply to the user message. While doing so
        # we also keep track of the number of tokens consumed.
        tokens_consumed_before = self.answer_model.total_tokens_consumed
        reply = self.answer_model.predict(self.messages.prompt(),
                                          attachments=attachments)
        if isinstance(reply, str) and self.documentation.poor_match:
            reply = '''<div class="message-info" markdown="1">Expert knowledge is enabled, but Sigmund was unable to find useful documentation to answer your question. To get a more useful answer:
            
- Provide more details and relevant keywords
- Or: Enable expert knowledge that is relevant to your question (see Menu)
- Or: Disable all expert knowledge to discuss general subjects (see Menu)
</div>\n\n''' + reply
        tokens_consumed = self.answer_model.total_tokens_consumed \
            - tokens_consumed_before
        logger.info(f'tokens consumed: {tokens_consumed}')
        self.database.add_activity(tokens_consumed)        
        if config.log_replies:
            logger.info(f'[{state} state] reply: {reply}')
        # If the reply is a callable, then it's a tool that we need to run
        if callable(reply):
            tool_message, tool_result, tool_language, needs_feedback = reply(
                attachments)
            if needs_feedback:
                logger.info(f'[{state} state] tools need feedback')
                if not self.answer_model.supports_tool_feedback:
                    logger.info(
                        f'[{state} state] model does not support feedback')
                    needs_feedback = False
            tool_message, workspace_content, workspace_language = \
                utils.extract_workspace(tool_message)
            # If the tool has a result and no workspace content was included in
            # the message itself, use the result as the workspace message. Also
            # remember the tool result it in the mssage history.
            if tool_result:
                if workspace_content is None:
                    workspace_content = tool_result['content']
                    workspace_language = tool_language
                metadata = self.messages.append(
                    'assistant', message=tool_message,
                    workspace_content=workspace_content,
                    workspace_language=workspace_language)
                self.messages.append('tool', message=json.dumps(tool_result))
            else:
                metadata = self.messages.append('assistant', tool_message)
            yield Reply(tool_message, metadata, workspace_content,
                        workspace_language)
        # Otherwise the reply is a regular AI message
        else:
            reply, workspace_content, workspace_language = \
                utils.extract_workspace(reply)
            metadata = self.messages.append(
                'assistant',
                message=reply,
                workspace_content=workspace_content,
                workspace_language=workspace_language)
            yield Reply(reply, metadata, workspace_content, workspace_language)
            needs_feedback = False
        # If feedback is required by a tool, go for another round.
        if needs_feedback and not self._rate_limit_exceeded():
            if workspace_content is not None:
                logger.info('workspace content has been updated for feedback')
                self.messages.workspace_content = workspace_content
                self.messages.workspace_language = workspace_language
            for reply in self._answer(attachments, state='feedback'):
                yield reply



import logging
import tempfile
import base64
import os
from .. import config, utils
from . import BaseModel
from ._openai_model import OpenAIModel
logger = logging.getLogger('sigmund')


class MistralModel(OpenAIModel):
    
    supports_not_done_yet = False

    def __init__(self, sigmund, model, **kwargs):
        from mistralai import Mistral
        BaseModel.__init__(self, sigmund, model, **kwargs)
        self._actual_model = self._model
        # Mistral doesn't allow a tool to be specified by name. So if this
        # happens, we instead use the 'any' option, which forces use of the
        # best fitting tool, which in the case of a single tool boils down to
        # the same thing as forcing the tool by name.
        if self._tool_choice not in [None, 'none', 'auto', 'any']:
            self._tool_choice = 'any'
        self._client = Mistral(api_key=config.mistral_api_key)
        
    def predict(self, messages, attachments=None, track_tokens=True):
        if isinstance(messages, str):
            messages = [self.convert_message(messages)]
        else:
            messages = utils.prepare_messages(messages, allow_ai_first=False,
                                              allow_ai_last=False,
                                              merge_consecutive=True)
            messages = [self.convert_message(message) for message in messages]
            messages = self._prepare_tool_messages(messages)
        # Mistral requires an assistant message after a tool message
        while True:
            for i, message in enumerate(messages[:-1]):
                next_message = messages[i + 1]
                if message['role'] == 'tool' and \
                        next_message['role'] == 'user':
                    break
            else:
                break
            logger.info('adding assistant message between tool and user')
            messages.insert(i + 1, {'role': 'assistant',
                                'content': 'Tool was executed.'})
        # Attachments are included with the last message. The content is now
        # no longer a single str, but a list of dict
        self._actual_model = self._model
        if attachments:
            logger.info('adding attachments to last message')
            content = [{'type': 'text', 'text': messages[-1]['content']}]
            for attachment in attachments:
                # Decompose the HTML-style data into a mimetype and the 
                # actual data
                url = attachment['url']
                data = url[url.find(',') + 1:]                
                if attachment['type'] == 'image':
                    # Vision requires a special model (pixtral). This model also
                    # understands text and documents, so whenever there is a 
                    # single attachment, we switch to using this model.
                    self._actual_model = \
                        config.model_config['mistral']['vision_model']
                    content.append({'type': 'image_url',
                                    'image_url': attachment['url']})
                elif attachment['type'] == 'document':
                    # Documents have to be uploaded first, and then provided as
                    # a url
                    tmp_file = tempfile.NamedTemporaryFile(delete=False)
                    tmp_file.write(base64.b64decode(data))
                    tmp_file.close()
                    uploaded_pdf = self._client.files.upload(
                        file={
                            "file_name": attachment['file_name'],
                            "content": open(tmp_file.name, 'rb'),
                        },
                        purpose="ocr"
                    )
                    os.remove(tmp_file.name)
                    signed_url = self._client.files.get_signed_url(file_id=uploaded_pdf.id)
                    content.append({'type': 'document_url',
                                    'document_url': signed_url.url})
            messages[-1]['content'] = content
        return BaseModel.predict(self, messages, attachments, track_tokens)
        
    def get_response(self, response) -> [str, callable]:
        content = response.choices[0].message.content
        tool_message_prefix = ''
        # During thinking, content consists of a mix of text and thinking 
        # blocks, where thinking blocks themselves consist of chunks of text.
        # For now, we simply concatenate everything into a single large 
        # response. This is because there may be multiple thinking and text 
        # blocks mixed in a single response, see also:
        # - <https://github.com/mistralai/client-python/issues/252>
        if isinstance(content, list):
            text = []
            tool_message_prefix = ''
            for block in content:
                if block.type == 'text':
                    text.append(block.text)
                    tool_message_prefix += block.text
                if block.type == 'thinking':
                    for thinking_chunk in block.thinking:
                        text.append(thinking_chunk.text)
                        tool_message_prefix += thinking_chunk.text
            content = '\n'.join(text)
        # If tool calls are present, we execute the tool using the current text
        # as a prefix.
        tool_calls = response.choices[0].message.tool_calls
        if tool_calls:
            function = tool_calls[0].function
            if self._tools:
                for tool in self._tools:
                    if tool.name == function.name:
                        return tool.bind(function.arguments,
                                         message_prefix=tool_message_prefix + '\n\n')
            logger.warning(f'invalid tool called: {function}')
            return self.invalid_tool            
        return content
        
    def _tool_call_id(self, nr):
        # Must be a-z, A-Z, 0-9, with a length of 9
        return f'{nr:09d}'
        
    def _mistral_invoke(self, fnc, messages):
        # Mistral tends to get stuck in a loop where the same tool is called
        # over and over again. To fix this, we temporarily disallow tools when
        # the last message was a tool.	
        if messages[-1]['role'] == 'tool':
            kwargs = {}
        else:
            kwargs = self._tool_args()        
        kwargs.update(config.mistral_kwargs)
        if self.json_mode:
            kwargs['response_format'] = {"type": "json_object"}
        return fnc(model=self._actual_model, messages=messages, **kwargs)
    
    def invoke(self, messages):
        return self._mistral_invoke(self._client.chat.complete, messages)
        
    def async_invoke(self, messages, attachments=None):
        return self._mistral_invoke(self._client.chat.complete_async, messages)



import jinja2

# The system prompt used during question answering is composed of the fragments
# below

SYSTEM_PROMPT_IDENTITY = '''You are Sigmund, a brilliant AI assistant. You always put code between triple backticks (```), and never use triple backticks for anything else. You sometimes use emojis.

When providing examples or updated text or code to the user always do this through the workspace. You set workspace content by including `<workspace language="language">text or code</workspace>` in your reply or by using tools.

Here is an example:

```
Sure, I can write a hello world function for you! I added the code to the workspace.

<workspace language="python">
def hello_world():
    print('Hello world!')
</workspace>
```

Available languages: css, html, javascript, opensesame, python, r, markdown

Important: always use the workspace as shown above, and do *not* include long examples of text or code in your reply.
'''

SYSTEM_PROMPT_CONDENSED = '''Below is a summary of the start of the conversation:

<summary>
{{ summary }}
</summary>'''

CONDENSE_HISTORY = '''Summarize the following conversation:

{{ history }}
'''

JUDGE_RELEVANCE = '''Is the document relevant for answering the question?

# Question

<question>
{{ question }}
</question>

# Document

<document>
{{ documentation }}
</document>

# Reply format

Reply with a JSON string to indicate whether the document is relevant ({"relevant": true}) or not ({"relevant": false}). Do not include any additional text.'''

TITLE_PROMPT = '''Provide a brief title that desribes the topic of the conversation below. Reply only with the title, do not add any additional text.'''
DESCRIBE_PROMPT = '''Provide a brief description of the following text:
        
Filename: {{ name }}
        
<TEXT>
{{ text_representation }}
</TEXT>'''
SUMMARIZE_PROMPT = '''Provide a long comprehensive summary of the following text:

<TEXT>
{{ text_representation }}
</TEXT>'''
PUBLIC_SEARCH_PROMPT = '''For each of the documentation sections below, provide a summary in a single bullet point using this format: "- [title](url): description (max 2 sentences)". Only reply with bullet points.
    
{{ documentation }}
'''

CURRENT_WORKSPACE = '''## Current workspace content

The workspace contains the following content:

<workspace language="{{ workspace_language}} ">
{{ workspace_content }}
</workspace>
'''


def render(tmpl, **kwargs):
    return jinja2.Template(tmpl).render(**kwargs)



import json
import base64
import mimetypes
import logging
from flask import request, jsonify, Response, redirect, session, \
    stream_with_context, make_response, Blueprint
from werkzeug.exceptions import RequestEntityTooLarge
from flask_login import login_required
from .. import config
from ..redis_client import redis_client
from .app import get_sigmund
logger = logging.getLogger('sigmund')
api_blueprint = Blueprint('api', __name__)


@api_blueprint.route('/health', methods=['GET'])
@login_required
def health():
    return jsonify(success=True)


@api_blueprint.route('/chat/start', methods=['POST'])
@login_required
def api_chat_start():

    # Store request data in the session
    try:
        message = request.form.get('message', '')
        workspace_content = request.form.get('workspace_content', '')
        workspace_language = request.form.get('workspace_language', '')
        message_id = request.form.get('message_id', '')
        transient_settings = request.form.get('transient_settings', '{}')
        transient_settings = json.loads(transient_settings)
        transient_system_prompt = request.form.get(
            'transient_system_prompt', '')
        foundation_document_topics = request.form.get(
            'foundation_document_topics', None)
    except RequestEntityTooLarge:
        logger.error('Request too large')
        return jsonify({'error': 'Request too large'}), 413
    if foundation_document_topics is not None:
        foundation_document_topics = json.loads(foundation_document_topics)
    session['user_message']       = message
    session['workspace_content']  = workspace_content
    session['workspace_language'] = workspace_language
    session['message_id']         = message_id
    session['transient_settings'] = transient_settings
    session['transient_system_prompt'] = transient_system_prompt
    session['foundation_document_topics'] = foundation_document_topics

    sigmund = get_sigmund(transient_settings=transient_settings)
    redis_client.delete(f'stream_cancel_{sigmund.user_id}')

    # Grab attached files from the request
    attachments = request.files.getlist('attachments')
    attachment_list = []
    for file in attachments:
        if not file:
            continue
        # Read file contents
        file_bytes = file.read()
        file_b64 = base64.b64encode(file_bytes).decode('utf-8')
        # Guess a MIME type for building the data URL
        mime_type, _ = mimetypes.guess_type(file.filename)
        if not mime_type:
            # Fallback if unable to guess the MIME type
            mime_type = 'application/octet-stream'
        # Build data URL
        data_url = f"data:{mime_type};base64,{file_b64}"
        # Define the type field
        # (Optional) Distinguish between documents and images more precisely:
        if mime_type.startswith('image/'):
            attachment_type = 'image'
        else:
            attachment_type = 'document'
        attachment_list.append({
            'type': attachment_type,
            'file_name': file.filename,
            'url': data_url
        })
    # Save to Redis
    # Convert list of attachments to JSON first
    redis_client.set(f'attachments_{sigmund.user_id}', json.dumps(attachment_list))
    return '{}'


@api_blueprint.route('/chat/stream', methods=['GET'])
@login_required
def api_chat_stream():

    # Retrieve message details from session
    message = session.get('user_message', '')
    workspace_content = session.get('workspace_content', '')
    workspace_language = session.get('workspace_language', '')
    message_id = session.get('message_id', '')
    transient_settings = session.get('transient_settings')
    transient_system_prompt = session.get('transient_system_prompt')
    foundation_document_topics = session.get('foundation_document_topics')

    # Retrieve attachments from Redis
    sigmund = get_sigmund(transient_settings=transient_settings,
                          transient_system_prompt=transient_system_prompt,
                          foundation_document_topics=foundation_document_topics)
    attachments_json = redis_client.get(f'attachments_{sigmund.user_id}')
    attachments = []
    if attachments_json:
        attachments = json.loads(attachments_json)

    logger.info(f'starting stream for {sigmund.user_id}')
    def generate():
        for reply in sigmund.send_user_message(
                message, workspace_content, workspace_language,
                attachments=attachments, message_id=message_id):
            json_reply = reply.to_json()
            logger.debug(f'ai message: {json_reply}')
            yield f'data: {json_reply}\n\n'
            if redis_client.get(f'stream_cancel_{sigmund.user_id}'):
                logger.info(f'stream cancelled for {sigmund.user_id}')
                break
        yield 'data: {"action": "close"}\n\n'

    # Return a server-sent event response
    return Response(stream_with_context(generate()),
                    mimetype='text/event-stream')


@api_blueprint.route('/chat/cancel', methods=['POST'])
@login_required
def api_chat_cancel_stream():
    sigmund = get_sigmund()
    logger.info(f'cancelling stream for {sigmund.user_id}')
    redis_client.set(f'stream_cancel_{sigmund.user_id}', '1')
    return jsonify({'status': 'cancelled'}), 200
    
    
@api_blueprint.route('/conversation/new')
@login_required
def new_conversation():
    sigmund = get_sigmund()
    sigmund.database.new_conversation()
    return redirect('/chat')


@api_blueprint.route('/conversation/clear')
@login_required
def clear_conversation():
    sigmund = get_sigmund()
    sigmund.messages.init_conversation()
    sigmund.messages.save()
    return redirect('/chat')


@api_blueprint.route('/conversation/activate/<int:conversation_id>',
                     methods=['GET', 'POST'])
@login_required
def activate_conversation(conversation_id):
    if conversation_id:
        sigmund = get_sigmund()
        sigmund.database.set_active_conversation(conversation_id)
    logger.info('redirecting to chat')
    return redirect('/chat')


@api_blueprint.route('/conversation/list', methods=['GET'])
@login_required
def list_conversations():
    sigmund = get_sigmund()
    query = request.args.get('query', None)
    return jsonify(sigmund.database.list_conversations(query))
    

@api_blueprint.route('/conversation/delete/<int:conversation_id>',
                     methods=['DELETE'])
@login_required
def delete_conversation(conversation_id):
    sigmund = get_sigmund()
    sigmund.database.delete_conversation(conversation_id)
    return '', 204


@api_blueprint.route('/conversation/export', methods=['GET'])
@login_required
def export_conversations():
    sigmund = get_sigmund()
    conversations = sigmund.database.export_conversations()
    if conversations:
        response = make_response(json.dumps(conversations))
        response.headers['Content-Type'] = 'application/json'
        response.headers['Content-Disposition'] = \
            'attachment; filename=sigmund-export.json'
        return response
    return jsonify(success=False,
                   message="No conversations found or access denied"), 404


@api_blueprint.route('/message/delete/<message_id>', methods=['DELETE'])
@login_required
def delete_message(message_id):
    sigmund = get_sigmund()
    sigmund.messages.delete(message_id)
    return jsonify(success=True)


@api_blueprint.route('/setting/set', methods=['POST'])
@login_required
def set_setting():
    data = request.json
    if not isinstance(data, dict):
        logger.warning(f'invalid setting data: {data}')
        return jsonify(success=False, message='invalid setting data')
    sigmund = get_sigmund()
    logger.info(f'setting: {data}')
    for key, value in data.items():
        sigmund.database.set_setting(key, value)
    return jsonify(success=True)
    

@api_blueprint.route('/setting/get/<key>')
@login_required
def get_setting(key):
    sigmund = get_sigmund()
    value = sigmund.database.get_setting(key)
    if value is None:
        if key in config.settings_default:
            value = config.settings_default[key]
            logger.info(f'using default setting for {key}: {value}')
        else:
            return jsonify(success=False, value=value)
    logger.info(f'get setting: {key} = {value}')
    return jsonify(success=True, value=value)
    



from . import BaseTool
from .. import config
import logging
import json
import base64
import pyalex
from pyalex import Works
from mistralai import Mistral

logger = logging.getLogger('sigmund')
pyalex.config.api_key = config.openalex_api_key


class search_openalex(BaseTool):
    """Search OpenAlex for scientific articles. An overview of metadata (abstracts, authors, titles, etc.) of matching articles will be returned, including OpenAlex IDs that can be used with download_from_openalex."""

    arguments = {
        "queries": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "description": "A list of search queries",
        },
        "max_results_per_query": {
            "type": "integer",
            "description": "The maximum number of results to return per query. (default=10)",
        }
    }
    required_arguments = ["queries"]

    def __call__(self, queries, max_results_per_query=10):
        results = []
        if isinstance(queries, str):
            queries = [queries]
        logger.info(f'searching OpenAlex for {len(queries)} queries')
        for query in queries:
            pager = (Works()
                     .search(query)
                     .paginate(
                         per_page=min(max_results_per_query, 25),
                         n_max=max_results_per_query))
            for page in pager:
                for work in page:
                    logger.info(
                        f'appending doc for OpenAlex search: {query}')
                    oa = work.get("open_access") or {}
                    location = work.get("primary_location") or {}
                    source = location.get("source") or {}
                    result = {
                        "openalex_id": work.get("id"),
                        "title": work.get("title"),
                        "doi": work.get("doi"),
                        "year": work.get("publication_year"),
                        "authors": [
                            a["author"]["display_name"]
                            for a in work.get("authorships", [])
                        ],
                        "abstract": work.get("abstract"),
                        "venue": source.get("display_name"),
                        "cited_by_count": work.get("cited_by_count"),
                        "is_open_access": oa.get("is_oa"),
                        "oa_url": oa.get("oa_url"),
                    }
                    results.append(result)
        query_list = '\n'.join(f'- {q}' for q in queries)
        message = (f'I searched for:\n\n{query_list}\n\nI found '
                   f'{len(results)} articles and added them to the workspace.')
        results = json.dumps(results, indent='  ')
        return message, results, True


class download_from_openalex(BaseTool):
    """Download the full text (in TEI-XML format) of a single scientific article from OpenAlex using its OpenAlex ID."""

    arguments = {
        "openalex_id": {
            "type": "string",
            "description": "The OpenAlex ID of the article (e.g. 'https://openalex.org/W1234567890' or 'W1234567890').",
        }
    }
    required_arguments = ["openalex_id"]

    def __call__(self, openalex_id):
        if not openalex_id.startswith("https://"):
            openalex_id = f"https://openalex.org/{openalex_id}"
        work = Works()[openalex_id]
        if not work['has_content']['pdf']:
            return f"No full text available for {openalex_id}.", "", True
        try:
            pdf_content = work.pdf.get()
        except Exception as e:
            logger.error(f"Failed to download PDF for {openalex_id}: {e}")
            return "Failed to download full text for this article.", "", True
        try:
            full_text = _ocr_extract(pdf_content, "application/pdf")
        except Exception as e:
            logger.error(f"Failed to convert PDF to text for {openalex_id}: {e}")
            return "Failed to convert PDF to text for this article.", "", True            
        title = work.get("title", openalex_id)
        message = (f'I downloaded the full text of "{title}" and added '
                   f'it to the workspace.')
        return message, full_text, True


def _ocr_extract(content, mimetype):
    """Extract text from a document using Mistral's OCR API.

    Parameters
    ----------
    content : bytes
        The document content.
    mimetype : str
        The MIME type of the document.

    Returns
    -------
    str
        The extracted text.
    """
    b64_content = base64.b64encode(content).decode('utf-8')    
    client = Mistral(api_key=config.mistral_api_key)
    ocr_response = client.ocr.process(
        model="mistral-ocr-latest",
        document={
            "type": "document_url",
            "document_url": f"data:{mimetype};base64,{b64_content}" 
        },
        include_image_base64=False
    )
    return '\n\n'.join(page.markdown for page in ocr_response.pages)


import json
from . import config, utils, process_sigmund_message


class BaseReply:
    
    def __contains__(self, needle):
        return needle.lower() in self.to_json().lower()
        
    def __str__(self):
        return self.to_json()
        
        
class Reply(BaseReply):
    """Corresponds to a regular AI reply that is sent to the client."""
    def __init__(self, msg: str, metadata: dict,
                 workspace_content: str=None, workspace_language: str='text'):
        self.msg = msg
        self.metadata = metadata
        self.workspace_content = workspace_content
        self.workspace_language = workspace_language
        
    def to_json(self):
        return json.dumps(
            {'response': utils.md(
                process_sigmund_message.process_ai_message(self.msg)),
             'metadata': self.metadata,
             'workspace_content': self.workspace_content,
             'workspace_language': self.workspace_language
            }
        )
        
        
class ActionReply(BaseReply):
    """Corresponds to an action reply that is sent to the client, typically to
    set the loading indicator.
    """
    def __init__(self, msg, action='set_loading_indicator'):
        self.msg = msg
        self.action = action
        
    def to_json(self):
        return json.dumps({'action': self.action, 'message': self.msg})
